from or_pcd.Preprocessor.Downsamplers.farthestDownsampler import FarthestDownsampler
from or_pcd.Preprocessor.Downsamplers.randomDownsampler import RandomDownsampler
from or_pcd.Preprocessor.Downsamplers.voxelDownsampler import VoxelDownsampler

